/*	Long64Lib.c

	(c) Osmose Editeur - www.osmose.net
	Thibaud Arguillere - targuillere@osmose.net

	January, 04, 2005
		Just recompile for mach-o and 4D 2004
		No new version number for this

	December, 24, 2002

	This plug-in brings 8 bytes long to 4D developers
*/

/********************** THIS IS FREE ! *****************************
	This is a Christmas gift for every 4D developer and every
	4D end-user.

	This plug-in comes for free and can be used without paying
	anything. As it is free, it also comes with no warranty, tech
	support, etc.
	-----------------> USE IT AT YOUR OWN RISK <-----------------
	
	If you want, you may incorporate this code into your own plug-ins
	or applications, with no restriction. If you re-distribute the
	source, we require that you make it clear in the source that
	the code was descended from Osmose Editeur.

	If you change the code, we'll be happy to know what you did,
	so feel free to send us your code at targuillere@osmose.net.
	Please note that:
		-> this is not mandatory, you can change the code and keep
		   the changes for your own use, even in a commercial
		   product,
		-> The code you will be plubished and everyone will be able
		   to get it for free, change it, etc... The source code
		   will state the you made some changes.

	We guess that the next major improvment will be the managment
	of _unsigned_ long long. But since this date, maybe
	4D GoldFinger will be there and this will be the end of
	this plug-in !

********************************************************************/

/*
	All the plug-in is, at the end, just a big wrapping
	of the native long64 implementation that one can find
	in almost every compiler since years.

	* How and where are the long64 stored ?
	---------------------------------------------------
	Our decision for the porting to 4D is to store the
	long64 in 4D's REAL type of data. In 4D, a real is
	nothing else than a double, that always use 8 bytes.
	[Oldest versions of 4D used 10 bytes real on 68k Mac]
	So the developper must store all his/her long64 in
	variables/fields explicitely declared as Real.
	NOTE : because of this (storing long64 in doubles),
	the developper _must_ avoid to manipulate the doubles
	themselves (see below) and _must_ always use the
	long64_New routine to create long64.

	* How to create a long64 ?
	---------------------------------------------------
	To create a new long64, use the long64_New function.
	This function can use 2 syntaxes:
		-> if you pass a non-empty string as its first
		   parameter, it uses it to build a long64. The
		   string must just have numbers, and can start
		   with "-" or "+". No space, nothing else.
				$aLong64:=long64_New("10 000 000 000 000") ` BAD
				$aLong64:=long64_New("10000000000000") ` OK

		-> if you pass an empty string as its first
		   parameter, then you must pass 2 longs as param
		   2 and 3. Param2 will be the low long of the long64,
		   and param3 will be the high long:
				$aLong:=long64_New("";0x11223344;0x55667788)
				=> this creates the long64 0x11223344556677
				
	* Manipulating long64
	---------------------------------------------------
	Because the plug-in stores long64 in doubles, you must
	_never_ use any basic operator on reals containing in fact
	long64. You _must_, instead, use all the accessors that
	are in the plug-in.

	As an example, if l1 and l2 are declared as Real but
	store in fact 2 long64, you can't use things such as
	(ILLEGAL means "you'll get wrong result"):
	
		$result:=$l1 + $l2  ` ILLEGAL. Use long64_Add
		$result:=$l1 - $l2  ` ILLEGAL. Use long64_Substract
		$result:=$l1 * $l2  ` ILLEGAL. Use long64_Multiply
		$result:=$l1 / $l2  ` ILLEGAL. Use long64_Divide
		$result:=$l1 % $l2  ` ILLEGAL. Use long64_Modulo
		
		$result:=$l1 & $l2  ` ILLEGAL. Use long64_AND
		$result:=$l1 | $l2  ` ILLEGAL. Use long64_InclusiveOR
		$result:=$l1 ^| $l2  ` ILLEGAL. Use long64_ExclusiveOR
		[long64_NOT realises the bitwise complement]
		$result:=$l1 >> 25 ` ILLEGAL. Use long64_ShiftRight
		$result:=$l1 << 25 ` ILLEGAL. Use long64_ShiftLeft

	To get low 4 bytes long or High 4 bytes long of a long64,
	use the long64_LowLong and long64_HighLong routines.

	This restriction also apply on comparison:
		if($l1 > $l2)  ` ILLEGAL : use long64_Compare

	If you want to add/Substract/Multiply/DIvide a numeric
	constant to a long64, you must build a long64 from the
	constant, using the long64_New routine:
		  ` Add 1 to $aLong64 :
		$aLong64:=long64_Add($aLong64; long64_New("1"))

	* Utilities
	---------------------------------------------------
	- long64_MainDispacth lets the developper use a more
	easy way the EXECUTE command.
	- long64_MAX returns the max. value of a long 64. This
	value is 0x7FFFFFFFFFFFFFFF.
	- long64_MIN returns the min. value of a long 64. This
	value is (-0x7FFFFFFFFFFFFFFF - 1).
	- long64_ByteSwap byte swaps the long64.

	* Syntax
	---------------------------------------------------
	ARITHMETIC OPERATIONS
	---------------------
		long64_Add
		long64_Substract
		long64_Multiply
		long64_Divide
		long64_Modulo
		
		Those routines have all the same syntax:
			result:=routine(l1;l2)

			Parameter	4D Type
				l1		real (containing a long64)
				l2		real (containing a long64)
				result	real (containing a long64)

		Those routines expect 2 parameters : a 4D real
		that has been filled with the long64_New routine
		(or are the result of a previous operation)
		They return the result of the operation. The names
		of the routines are, in our opinion, so clear that
		they don't need more explanations ;->

			result:=long64_Add(aLong64;anOtherOne)
		
	Bits manipulation
	---------------------
		long64_AND
		long64_InclusiveOR
		long64_ExclusiveOR
		long64_NOT
		long64_ShiftRight
		long64_ShiftLeft
		long64_HighLong
		long64_LowLong

		long64_AND, long64_InclusiveOR and long64_ExclusiveOR
		have the following syntax:
			result:=routine(l1;l2)

			Parameter	4D Type
				l1		real (containing a long64)
				l2		real (containing a long64)
				result	real (containing a long64)

		l1 and l2 must have been created by long64_New
		of are the result of a previous long64_nn operation.

		long64_NOT expects only one parameter (a 4D real)
		and return its bitwise complement:
			result:=long64_NOT(aLong64)

		long64_ShiftRight and long64_ShiftLeft do what they
		are supposed to do. The parameters are:
			Parameter	4D Type
				l1		real (containing a long64)
				l2		Integer or long integer
				result	real (containing a long64)
		
			IMPORTANT : the plug-in does not check the bit
			number to see if it is >= 0 and <= 63.

		long64_HighLong and long64_lowLong return the
		high 4 bytes and the low 4 bytes of the long64.
		They return a value of king Long integer.

	long64_Compare
	---------------------
			result:=long64_Compare(l1;l2)
	As long64 are stored in reals, you can't compare
	directly 2 reals storing, in fact, 2 long64. You
	must use this function that expect 2 4D reals and
	returns a short intreger value that will be:
			0 : l1 = l2
			-1 : l1 < l2
			1 : l1 > l2

	UTILITIES and MISC.
	---------------------
	long64_New creates a new long64 value and returns it
	in a 4D real. See at the beginning of this explanation.

	long64_ToString
		$l64AsString:=long64_ToString(aLong64)
		aLong64 is a long64 built with long64_New or is the
		result of a previous operation.
		l64AsString is aLong64 bconverted to a string. There
		is no formatting of the string. As an example :
			$strMax:=long64_ToString long64_MAX)
		...puts "9223372036854775807" in $strMax.

	long64_MainDispatch
		$result:=long64_MainDispatch(selector;l1;l2)

			Parameter		4D Type
				selector	string of 1
				l1			real (containing a long64)
				l2			real (containing a long64)
				result		real (containing a long64)

		Selector is a string/text. Only its first char
		is tested. It must contain the symbol of a basic
		operation :
				"+", "-", "*", "/", "%"
				"1", "|", "^", "~"
			For Shift left or right, use on single sign
			instead of 2: ">" or "<"

	long64_MAX has no parameter and returns the maximum
		value of a long64 : 0x7FFFFFFFFFFFFFFF
							9 223 372 036 854 775 807

	long64_MIN has no parameter and returns the maximum
		value of a long64 : -0x7FFFFFFFFFFFFFFF - 1
							-9 223 372 036 854 775 808

	long64_ByteSwap expects one parameter (a 4D real
	cointaining in fact a long 64) and returns a 4D
	real containing the parameter byte-swapped. This
	routinbe may be usefull if the developper exchanges
	some long64 between Mac/PC.
*/

// After this checking, we don't have to care
// that much when testing VERSIONWIN and VERSIONMAC :
// it'll always be Mac or Win.
// We need this since maybe other compilers use different
// naming for long64, including different format for prinf.
// If you compile for Windows with MetroWerks CodeWarrior, you
// should make little changes here and there: search for
// "VERSIONWIN" and "VERSIONMAC" in the code.
#if ( !_WIN32 && !_MSC_VER && !__POWERPC__ && !__MWERKS__ )
#	error "Must be (Win and VC++) or (Mac and CodeWarrior), please"
#endif

/*
	==================================================
	Includes
	==================================================
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "4DPluginAPI.h"
#include "Long64Lib.h"
					
/*
	==================================================
	typedefs
	==================================================
*/
#ifndef PA_ulong64
#	if	VERSIONWIN
	typedef unsigned __int64	PA_ulong64;
#	else
	typedef unsigned long long	PA_ulong64;
#	endif
#endif

/*
	==================================================
	Macros
	==================================================
*/
/*
	We try to improve a little bit the speed. The cost
	of the speed improvement is this one : a bit less
	readibility of the code, that is more difficult to
	debug because we use macros to avoid function calls.
	
	On this particular plug-in, this should not be a
	problem.
*/
#define RETURN_LONG64(_M_params, _M_L64)	\
				(**(PA_long64**) _M_params->fResult = ((PA_long64) _M_L64))

#define GET_LONG64_PARAMETER(_M_params, _M_index)	\
				(*( ( (PA_long64**) _M_params->fParameters ) [ _M_index - 1 ] ))

#define LONG_TO_LONG64(_M_Long)	( ((PA_long64)(long)_M_Long) & 0x00000000FFFFFFFF )

#define LONG64_MAX		0x7FFFFFFFFFFFFFFF
#define LONG64_MIN		((-LONG64_MAX) - 1)

#define LONG64_SWAP(_M_L64)		\
					( ((((PA_ulong64)_M_L64)<<56) & ((PA_ulong64) 0xFF00000000000000))| \
					((((PA_ulong64)_M_L64)<<40) & ((PA_ulong64) 0x00FF000000000000))  | \
					((((PA_ulong64)_M_L64)<<24) & ((PA_ulong64) 0x0000FF0000000000))  | \
					((((PA_ulong64)_M_L64)<< 8) & ((PA_ulong64) 0x000000FF00000000))  | \
					((((PA_ulong64)_M_L64)>> 8) & ((PA_ulong64) 0x00000000FF000000))  | \
					((((PA_ulong64)_M_L64)>>24) & ((PA_ulong64) 0x0000000000FF0000))  | \
					((((PA_ulong64)_M_L64)>>40) & ((PA_ulong64) 0x000000000000FF00))  | \
					((((PA_ulong64)_M_L64)>>56) & ((PA_ulong64) 0x00000000000000FF)) )

// This could be changed when using other compilers
#if VERSIONWIN
#	define	SPRINTF_LONG64_d	"%I64d"
#	define	SPRINTF_LONG64_X	"%I64X"
#elif VERSIONMAC
#	define	SPRINTF_LONG64_d	"%lld"
#	define	SPRINTF_LONG64_X	"%llX"
#endif
					
/*
	==================================================
	Misc.
	==================================================
*/
// _____________________________ 4D Plugin API wrapping
// This one could be, in fact, inlined. We put it in a Macro
/*
PA_long64 myPA_GetLong64Parameter( PA_PluginParameters params, short index );
PA_long64 myPA_GetLong64Parameter( PA_PluginParameters params, short index )
{
	return *( ( (PA_long64**) params->fParameters ) [ index - 1 ] );
}
*/
					
/*
	==================================================
	Code
	==================================================
*/
/*	PluginMain *************************************

*/
void PluginMain( long selector, PA_PluginParameters params )
{
	switch( selector )
	{
	case kInitPlugin:
	case kServerInitPlugin:
		PA_UseCStrings();
		break;

	case eCMD_long64_Add:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						+ GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_Substract:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						- GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_Multiply:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						* GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_Divide:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						/ GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_Modulo:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						% GET_LONG64_PARAMETER(params, 2) );
		break;

	case eCMD_long64_New:
		long64_New( params );
		break;
	case eCMD_long64_ToString:
		long64_ToString( params );
		break;
	case eCMD_long64_Compare:
		long64_Compare( params );
		break;
	case eCMD_long64_MainDispatch:
		long64_MainDispatch( params );
		break;

	case eCMD_long64_AND:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						& GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_InclusiveOR:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						| GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_ExclusiveOR:
		RETURN_LONG64(params,
						GET_LONG64_PARAMETER(params, 1)
						^ GET_LONG64_PARAMETER(params, 2) );
		break;
	case eCMD_long64_NOT:
		RETURN_LONG64(params,
						~GET_LONG64_PARAMETER(params, 1) );
		break;
	case eCMD_long64_ShiftRight:
		RETURN_LONG64(params,
						(GET_LONG64_PARAMETER(params, 1)
						>> PA_GetShortParameter(params, 2)) );
		break;
	case eCMD_long64_ShiftLeft:
		RETURN_LONG64(params,
						(GET_LONG64_PARAMETER(params, 1)
						<< PA_GetShortParameter(params, 2)) );
		break;
	case eCMD_long64_LowLong:
		PA_ReturnLong(params,
					(long) (GET_LONG64_PARAMETER(params, 1)
							& 0x00000000FFFFFFFF) );
		break;
	case eCMD_long64_HighLong:
		PA_ReturnLong(params,
					(long) ( (GET_LONG64_PARAMETER(params, 1)
							& 0xFFFFFFFF00000000) >> 32 ));
		break;

	case eCMD_long64_MAX:
		RETURN_LONG64(params, LONG64_MAX );
		break;
	case eCMD_long64_MIN:
		RETURN_LONG64(params, LONG64_MIN );
		break;

	case eCMD_long64_ByteSwap:
		RETURN_LONG64(params,
						LONG64_SWAP( GET_LONG64_PARAMETER(params, 1) ));
		break;
	}
} /* PluginMain */

/*	long64_ToString ******************************

*/
void long64_ToString( PA_PluginParameters params )
{
	char	returnValue[256];
	
	sprintf(returnValue,
			SPRINTF_LONG64_d,
			GET_LONG64_PARAMETER(params, 1));
	
	PA_ReturnString( params, returnValue );

} /* long64_ToString */

/*	long64_New ******************************

	$long64:=long64_New(string;lowLong;highLong)
	If string is "", we use lowLong and highLong to
	build the long64. Else, the string is converted.
*/
void long64_New( PA_PluginParameters params )
{
	char	long64AsString[256];

	PA_GetStringParameter( params, 1, long64AsString );

	if(long64AsString[0] != 0)
	{
#if (VERSIONWIN && _MSC_VER)
		RETURN_LONG64(params,
						_atoi64(long64AsString) );
#elif (VERSIONMAC && __MWERKS__)
		RETURN_LONG64(params,
						strtoll(long64AsString, NULL, 10) );
#else
#	error "Must be (Win && VC++) of (Mac && Metrowerks)"
#endif
	}
	else
	{
		PA_long64	lowLong, highLong;

		lowLong = LONG_TO_LONG64( PA_GetLongParameter(params, 2) );
		highLong = LONG_TO_LONG64( PA_GetLongParameter(params, 3) );

		RETURN_LONG64(params,
						(highLong << 32) | lowLong );
	}

} /* long64_New */

/*	long64_Compare ******************************

*/
void long64_Compare( PA_PluginParameters params )
{
	PA_long64	param1;
	PA_long64	param2;
	short		returnValue;

	param1 = GET_LONG64_PARAMETER( params, 1 );
	param2 = GET_LONG64_PARAMETER( params, 2 );

	if(param1 == param2)
		returnValue = 0;
	else if(param1 < param2)
		returnValue = -1;
	else
		returnValue = 1;

	PA_ReturnShort( params, returnValue );

} /* long64_Compare */

/*	long64_MainDispatch ******************************

*/
void long64_MainDispatch( PA_PluginParameters params )
{
	char		selector[256];
	PA_long64	param1;
	PA_long64	param2;
	PA_long64	resultLong64 = 0;

	PA_GetStringParameter( params, 1, selector );
	param1 = GET_LONG64_PARAMETER( params, 2 );
	param2 = GET_LONG64_PARAMETER( params, 3 );

	switch(selector[0])
	{
	case '+':
		resultLong64 = param1 + param2;
		break;
	case '-':
		resultLong64 = param1 - param2;
		break;
	case '*':
		resultLong64 = param1 * param2;
		break;
	case '/':
		resultLong64 = param1 / param2;
		break;
	case '%':
		resultLong64 = param1 % param2;
		break;
		
	case '&':
		resultLong64 = param1 & param2;
		break;
	case '|':
		resultLong64 = param1 | param2;
		break;
	case '^':
		resultLong64 = param1 ^ param2;
		break;
	case '~':
		resultLong64 = ~param1;
		break;
	case '>':
		resultLong64 = (param1 >> ((short)param2));
		break;
	case '<':
		resultLong64 = (param1 << ((short)param2));
		break;

	default:
		resultLong64 = 0;
		break;
	}

	RETURN_LONG64(params,
					resultLong64 );

} /* long64_MainDispatch */





